@@ -1,6 +1,7 @@
 /* GIO - GLib Input, Output and Streaming Library
  * 
- * Copyright (C) 2008 Benjamin Otte <otte@gnome.org>
+ * Copyright (C) 2008,2009 Benjamin Otte <otte@gnome.org>
+ *               2008,2009 Andreas Henriksson <andreas@fatal.se>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,7 +18,8 @@
  * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  * Boston, MA 02111-1307, USA.
  *
- * Author: Benjmain Otte <otte@gnome.org>
+ * Author: Benjamin Otte <otte@gnome.org>
+ *         Andreas Henriksson <andreas@fatal.se>
  */
 
 
@@ -96,7 +98,8 @@ typedef enum {
   FTP_FEATURE_MDTM = (1 << 0),
   FTP_FEATURE_SIZE = (1 << 1),
   FTP_FEATURE_TVFS = (1 << 2),
-  FTP_FEATURE_EPSV = (1 << 3)
+  FTP_FEATURE_EPSV = (1 << 3),
+  FTP_FEATURE_UTF8 = (1 << 4),
 } FtpFeatures;
 #define FTP_FEATURES_DEFAULT (FTP_FEATURE_EPSV)
 
@@ -106,6 +109,19 @@ typedef enum {
   FTP_SYSTEM_WINDOWS
 } FtpSystem;
 
+typedef enum {
+  /* Server advertises support for EPSV (or we assume that it supports it),
+   * but it does fail to do so, we set this flag so we can fall back to 
+   * PASV. */
+  FTP_WORKAROUND_BROKEN_EPSV = (1 << 0),
+  /* Server replies with a wrong address in PASV, we use connection IP 
+   * instead */
+  FTP_WORKAROUND_PASV_ADDR = (1 << 1),
+  /* server does not allow querying features before login, so we try after
+   * logging in instead. */
+  FTP_WORKAROUND_FEAT_AFTER_LOGIN = (1 << 2),
+} FtpWorkarounds;
+
 struct _GVfsBackendFtp
 {
   GVfsBackend		backend;
@@ -146,6 +162,7 @@ struct _FtpConnection
 
   FtpFeatures		features;
   FtpSystem		system;
+  FtpWorkarounds	workarounds;
 
   SoupSocket *		commands;
   gchar *	      	read_buffer;
@@ -169,6 +186,7 @@ ftp_connection_free (FtpConnection *conn)
 }
 
 #define ftp_connection_in_error(conn) ((conn)->error != NULL)
+#define ftp_connection_clear_error(conn) (g_clear_error (&(conn)->error))
 
 static gboolean
 ftp_connection_pop_job (FtpConnection *conn)
@@ -330,7 +348,6 @@ ftp_connection_receive (FtpConnection *conn,
 			ResponseFlags  flags)
 {
   SoupSocketIOStatus status;
-  gsize n_bytes;
   gboolean got_boundary;
   char *last_line;
   enum {
@@ -348,63 +365,62 @@ ftp_connection_receive (FtpConnection *conn,
   conn->read_bytes = 0;
   while (reply_state != DONE)
     {
-      if (conn->read_buffer_size - conn->read_bytes < 128)
-	{
-	  gsize new_size = conn->read_buffer_size + 1024;
-	  /* FIXME: upper limit for size? */
-	  gchar *new = g_try_realloc (conn->read_buffer, new_size);
-	  if (new)
-	    {
-	      conn->read_buffer = new;
-	      conn->read_buffer_size = new_size;
-	    }
-	  else
-	    {
-	      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-				   _("Invalid reply"));
-	      return 0;
-	    }
-	}
       last_line = conn->read_buffer + conn->read_bytes;
-      status = soup_socket_read_until (conn->commands,
-				       last_line,
-				       /* -1 byte for nul-termination */
-				       conn->read_buffer_size - conn->read_bytes - 1,
-				       "\r\n",
-				       2,
-				       &n_bytes,
-				       &got_boundary,
-				       conn->job->cancellable,
-				       &conn->error);
+      do {
+        gsize n_bytes;
 
-      conn->read_bytes += n_bytes;
-      conn->read_buffer[conn->read_bytes] = 0;
-      DEBUG ("<-- %s", last_line);
+        /* the available size must at least allow for boundary size (2)
+         * bytes to be available for reading */
+        if (conn->read_buffer_size - conn->read_bytes < 128)
+          {
+            gsize new_size = conn->read_buffer_size + 1024;
+            /* FIXME: upper limit for size? */
+            gchar *new = g_try_realloc (conn->read_buffer, new_size);
+            if (new)
+              {
+                /* make last line relative to new allocation */
+                last_line = new + (last_line - conn->read_buffer);
+                conn->read_buffer = new;
+                conn->read_buffer_size = new_size;
+              }
+            else
+              {
+                g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                                     _("Invalid reply"));
+                return 0;
+              }
+          }
+        status = soup_socket_read_until (conn->commands,
+                                         conn->read_buffer + conn->read_bytes,
+                                         /* -1 byte for nul-termination */
+                                         conn->read_buffer_size - conn->read_bytes - 1,
+                                         "\r\n",
+                                         2,
+                                         &n_bytes,
+                                         &got_boundary,
+                                         conn->job->cancellable,
+                                         &conn->error);
 
-      switch (status)
-	{
-	  case SOUP_SOCKET_OK:
-	  case SOUP_SOCKET_EOF:
-	    if (got_boundary)
-	      break;
-	    if (n_bytes > 0)
-	      continue;
+        conn->read_bytes += n_bytes;
+        conn->read_buffer[conn->read_bytes] = 0;
+
+        g_assert (status != SOUP_SOCKET_WOULD_BLOCK);
+        if (status == SOUP_SOCKET_ERROR)
+          return 0;
+
+        if (n_bytes == 0)
+          {
 	    g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
 				 _("Invalid reply"));
-	    /* fall through */
-	  case SOUP_SOCKET_ERROR:
-	    conn->read_buffer[conn->read_bytes] = 0;
-	    return 0;
-	  case SOUP_SOCKET_WOULD_BLOCK:
-	  default:
-	    g_assert_not_reached ();
-	    break;
-	}
+            return 0;
+          }
+      } while (!got_boundary);
+
+      DEBUG ("<-- %s", last_line);
 
       if (reply_state == FIRST_LINE)
 	{
-	  if (n_bytes < 4 ||
-	      last_line[0] <= '0' || last_line[0] > '5' ||
+	  if (last_line[0] <= '0' || last_line[0] > '5' ||
 	      last_line[1] < '0' || last_line[1] > '9' ||
 	      last_line[2] < '0' || last_line[2] > '9')
 	    {
@@ -428,8 +444,9 @@ ftp_connection_receive (FtpConnection *conn,
 	}
       else
 	{
-	  if (n_bytes >= 4 &&
-	      memcmp (conn->read_buffer, last_line, 3) == 0 &&
+	  if (last_line[0] == conn->read_buffer[0] &&
+              last_line[1] == conn->read_buffer[1] &&
+              last_line[2] == conn->read_buffer[2] &&
 	      last_line[3] == ' ')
 	    reply_state = DONE;
 	}
@@ -628,19 +645,36 @@ ftp_connection_parse_features (FtpConnection *conn)
     { "MDTM", FTP_FEATURE_MDTM },
     { "SIZE", FTP_FEATURE_SIZE },
     { "TVFS", FTP_FEATURE_TVFS },
-    { "EPSV", FTP_FEATURE_EPSV }
+    { "EPSV", FTP_FEATURE_EPSV },
+    { "UTF8", FTP_FEATURE_UTF8 },
   };
   char **supported;
   guint i, j;
 
-  supported = g_strsplit (conn->read_buffer, "\r\n", -1);
+  /* The "\n" should really be "\r\n" but we deal with both to handle broken
+   * servers. We strip off '\r' later, if it exists. */
+  supported = g_strsplit (conn->read_buffer, "\n", -1);
 
   for (i = 1; supported[i]; i++)
     {
-      const char *feature = supported[i];
+      char *feature = supported[i];
+      int len;
+
       if (feature[0] != ' ')
 	continue;
       feature++;
+
+      /* There should just be one space according to RFC2389, but some
+       * servers have more so we deal with any number of leading spaces.
+       */
+      while (g_ascii_isspace (feature[0]))
+        feature++;
+
+      /* strip off trailing '\r', if it exists. */
+      len = strlen(feature);
+      if (len > 0 && feature[len-1] == '\r')
+	      feature[len-1] = '\0';
+
       for (j = 0; j < G_N_ELEMENTS (features); j++)
 	{
 	  if (g_ascii_strcasecmp (feature, features[j].name) == 0)
@@ -740,6 +774,23 @@ ftp_connection_parse_system (FtpConnection *conn)
     }
 }
 
+static void
+ftp_connection_prepare (FtpConnection *conn)
+{
+  /* check supported features */
+  if (ftp_connection_send (conn, 0, "FEAT") != 0) {
+    ftp_connection_parse_features (conn);
+  } else {
+    ftp_connection_clear_error(conn);
+    conn->workarounds |= FTP_WORKAROUND_FEAT_AFTER_LOGIN;
+    conn->features = 0;
+  }
+
+  /* instruct server that we'll give and assume we get utf8 */
+  if (conn->features & FTP_FEATURE_UTF8)
+    ftp_connection_send (conn, 0, "OPTS UTF8 ON");
+}
+
 static gboolean
 ftp_connection_use (FtpConnection *conn)
 {
@@ -748,80 +799,44 @@ ftp_connection_use (FtpConnection *conn)
   if (ftp_connection_in_error (conn))
     return FALSE;
 
-  /* check supported features */
-  if (ftp_connection_send (conn, 0, "FEAT") != 0)
-    ftp_connection_parse_features (conn);
-  else
-    conn->features = FTP_FEATURES_DEFAULT;
-
+#if 0
   /* RFC 2428 suggests to send this to make NAT routers happy */
+  /* XXX: Disabled for the following reasons:
+   * - most ftp clients don't use it
+   * - lots of broken ftp servers can't see the difference between 
+   *   "EPSV" and "EPSV ALL"
+   * - impossible to dynamically fall back to regular PASV in case
+   *   EPSV doesn't work for some reason.
+   * If this makes your ftp connection fail, please file a bug and we will
+   * try to invent a way to make this all work. Until then, we'll just 
+   * ignore the RFC.
+   */
   if (conn->features & FTP_FEATURE_EPSV)
     ftp_connection_send (conn, 0, "EPSV ALL");
-  g_clear_error (&conn->error);
+  ftp_connection_clear_error(conn);
+#endif
+
+  if (conn->workarounds & FTP_WORKAROUND_FEAT_AFTER_LOGIN) {
+    if (ftp_connection_send (conn, 0, "FEAT") != 0) {
+      ftp_connection_parse_features (conn);
+    } else {
+      ftp_connection_clear_error(conn);
+      conn->features = FTP_FEATURE_EPSV;
+    }
+  }
 
   if (ftp_connection_send (conn, 0, "SYST"))
     ftp_connection_parse_system (conn);
-  g_clear_error (&conn->error);
+  ftp_connection_clear_error(conn);
 
   return TRUE;
 }
 
 static gboolean
-ftp_connection_ensure_data_connection (FtpConnection *conn)
+ftp_connection_open_data_connection (FtpConnection *conn, SoupAddress *addr)
 {
-  guint ip1, ip2, ip3, ip4, port1, port2;
-  SoupAddress *addr;
-  const char *s;
-  char *ip;
   guint status;
 
-  if (conn->features & FTP_FEATURE_EPSV)
-    {
-      status = ftp_connection_send (conn, RESPONSE_PASS_500, "EPSV");
-      if (STATUS_GROUP (status) == 2)
-	{
-	  s = strrchr (conn->read_buffer, '(');
-	  if (s)
-	    {
-	      guint port;
-	      s += 4;
-	      port = strtoul (s, NULL, 10);
-	      if (port != 0)
-		{
-		  addr = soup_address_new (
-		      soup_address_get_name (soup_socket_get_remote_address (conn->commands)),
-		      port);
-		  goto have_address;
-		}
-	    }
-	}
-    }
-  /* only binary transfers please */
-  status = ftp_connection_send (conn, 0, "PASV");
-  if (status == 0)
-    return FALSE;
-
-  /* parse response and try to find the address to connect to.
-   * This code does the sameas curl.
-   */
-  for (s = conn->read_buffer; *s; s++)
-    {
-      if (sscanf (s, "%u,%u,%u,%u,%u,%u", 
-                 &ip1, &ip2, &ip3, &ip4, 
-                 &port1, &port2) == 6)
-       break;
-    }
-  if (*s == 0)
-    {
-      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-			   _("Invalid reply"));
-      return FALSE;
-    }
-  ip = g_strdup_printf ("%u.%u.%u.%u", ip1, ip2, ip3, ip4);
-  addr = soup_address_new (ip, port1 << 8 | port2);
-  g_free (ip);
-
-have_address:
   conn->data = soup_socket_new ("non-blocking", FALSE,
 				"remote-address", addr,
 				"timeout", TIMEOUT_IN_SECONDS,
@@ -843,6 +858,122 @@ have_address:
   return TRUE;
 }
 
+static gboolean
+ftp_connection_ensure_data_connection_epsv (FtpConnection *conn)
+{
+  const char *s;
+  guint port;
+  SoupAddress *addr;
+  guint status;
+  gboolean connected;
+
+  if ((conn->features & FTP_FEATURE_EPSV) == 0)
+    return FALSE;
+
+  if (conn->workarounds & FTP_WORKAROUND_BROKEN_EPSV)
+    return FALSE;
+
+  status = ftp_connection_send (conn, RESPONSE_PASS_500, "EPSV");
+  if (STATUS_GROUP (status) != 2)
+    return FALSE;
+
+  s = strrchr (conn->read_buffer, '(');
+  if (!s)
+    return FALSE;
+
+  s += 4;
+  port = strtoul (s, NULL, 10);
+  if (port == 0)
+    return FALSE;
+
+  addr = soup_address_new (
+		      soup_address_get_name (soup_socket_get_remote_address (conn->commands)),
+		      port);
+
+  connected = ftp_connection_open_data_connection (conn, addr);
+  if (!connected)
+    {
+      DEBUG ("Successful EPSV response code, but data connection failed. Enabling FTP_WORKAROUND_BROKEN_EPSV.\n");
+      conn->workarounds |= FTP_WORKAROUND_BROKEN_EPSV;
+      ftp_connection_clear_error (conn);
+    }
+  return connected;
+}
+
+static gboolean
+ftp_connection_ensure_data_connection_pasv (FtpConnection *conn)
+{
+  guint ip1, ip2, ip3, ip4, port1, port2;
+  const char *s;
+  SoupAddress *addr;
+  guint status;
+
+  /* only binary transfers please */
+  status = ftp_connection_send (conn, 0, "PASV");
+  if (status == 0)
+    return FALSE;
+
+  /* parse response and try to find the address to connect to.
+   * This code does the same as curl.
+   */
+  for (s = conn->read_buffer; *s; s++)
+    {
+      if (sscanf (s, "%u,%u,%u,%u,%u,%u", 
+		 &ip1, &ip2, &ip3, &ip4, 
+		 &port1, &port2) == 6)
+       break;
+    }
+  if (*s == 0)
+    {
+      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+			   _("Invalid reply"));
+      return FALSE;
+    }
+
+  if (!(conn->workarounds & FTP_WORKAROUND_PASV_ADDR))
+    {
+      char *ip;
+
+      ip = g_strdup_printf ("%u.%u.%u.%u", ip1, ip2, ip3, ip4);
+      addr = soup_address_new (ip, port1 << 8 | port2);
+      g_free (ip);
+
+      if (ftp_connection_open_data_connection (conn, addr))
+        return TRUE;
+         
+      /* set workaround flag (see below), so we don't try this again */
+      DEBUG ("Successfull PASV response but data connection failed. Enabling FTP_WORKAROUND_PASV_ADDR.\n");
+      conn->workarounds |= FTP_WORKAROUND_PASV_ADDR;
+      ftp_connection_clear_error (conn);
+    }
+
+  /* Workaround code:
+   * Various ftp servers aren;t setup correctly when behind a NAT. They report
+   * their own IP address (like 10.0.0.4) and not the address in front of the
+   * NAT. But this is likely the same address that we connected to with our
+   * command connetion. So if the address given by PASV fails, we fall back 
+   * to the address of the command stream.
+   */
+  addr = soup_address_new (soup_address_get_name (soup_socket_get_remote_address (conn->commands)),
+			   port1 << 8 | port2);
+  return ftp_connection_open_data_connection (conn, addr);
+}
+
+static gboolean
+ftp_connection_ensure_data_connection (FtpConnection *conn)
+{
+  if (ftp_connection_ensure_data_connection_epsv (conn))
+    return TRUE;
+
+  if (ftp_connection_in_error (conn))
+    return FALSE;
+
+  if (ftp_connection_ensure_data_connection_pasv (conn))
+    return TRUE;
+
+  return FALSE;
+}
+
 static void
 ftp_connection_close_data_connection (FtpConnection *conn)
 {
@@ -929,7 +1060,7 @@ ftp_connection_try_cd (FtpConnection *conn, const FtpFile *file)
 
   if (!ftp_connection_cd (conn, file))
     {
-      g_clear_error (&conn->error);
+      ftp_connection_clear_error (conn);
       return FALSE;
     }
   
@@ -946,7 +1077,7 @@ dir_default_init_data (FtpConnection *conn, const FtpFile *dir)
 
   ftp_connection_send (conn,
 		       RESPONSE_PASS_100 | RESPONSE_FAIL_200,
-		       "LIST");
+		       (conn->system == FTP_SYSTEM_UNIX) ? "LIST -a" : "LIST");
 }
 
 static GFileInfo *
@@ -979,7 +1110,7 @@ dir_default_get_root (FtpConnection *conn)
 static gpointer
 dir_default_iter_new (FtpConnection *conn)
 {
-  return g_slice_new (struct list_state);
+  return g_slice_new0 (struct list_state);
 }
 
 static GFileInfo *
@@ -1062,9 +1193,15 @@ dir_default_iter_process (gpointer        iter,
 	      char *end = str + 4;
 	      char *start;
 	      start = str - 1;
-	      while (*start != '/')
+	      while (start >= symlink_file && *start != '/')
 		start--;
-	      memcpy (start + 1, end, strlen (end) + 1);
+
+	      if (start < symlink_file) {
+		      *symlink_file = '/';
+		      start = symlink_file;
+	      }
+
+	      memmove (start + 1, end, strlen (end) + 1);
 	    }
 	  str = symlink_file + strlen (symlink_file) - 1;
 	  while (*str == '/' && str > symlink_file)
@@ -1090,6 +1227,13 @@ dir_default_iter_process (gpointer        iter,
   g_free (s);
   g_free (name);
 
+  /* Workaround:
+   * result.fetime.tm_year contains actual year instead of offset-from-1900,
+   * which mktime expects.
+   */
+  if (result.fe_time.tm_year >= 1900)
+	  result.fe_time.tm_year -= 1900;
+
   tv.tv_sec = mktime (&result.fe_time);
   if (tv.tv_sec != -1)
     g_file_info_set_modification_time (info, &tv);
@@ -1159,39 +1303,48 @@ g_vfs_backend_ftp_pop_connection (GVfsBackendFtp *ftp,
 	break;
       conn = g_queue_pop_head (ftp->queue);
 
-      if (conn != NULL) {
-	/* Figure out if this connection had a timeout sent. If so, skip it. */
-	g_mutex_unlock (ftp->mutex);
-	ftp_connection_push_job (conn, job);
-	if (ftp_connection_send (conn, 0, "NOOP"))
-	  break;
-	    
-	g_clear_error (&conn->error);
-	conn->job = NULL;
-	ftp_connection_free (conn);
-	conn = NULL;
-	g_mutex_lock (ftp->mutex);
-	ftp->connections--;
-	continue;
-      }
-
-      if (ftp->connections < ftp->max_connections)
+      if (conn != NULL)
 	{
-	  ftp->connections++;
+	  /* Figure out if this connection had a timeout sent. If so, skip it. */
 	  g_mutex_unlock (ftp->mutex);
-	  conn = ftp_connection_create (ftp->addr, job);
-	  ftp_connection_login (conn, ftp->user, ftp->password);
-	  ftp_connection_use (conn);
-	  if (!ftp_connection_in_error (conn))
+	  ftp_connection_push_job (conn, job);
+	  if (ftp_connection_send (conn, 0, "NOOP"))
 	    break;
-
-	  ftp_connection_pop_job (conn);
+	      
+	  ftp_connection_clear_error (conn);
+	  conn->job = NULL;
 	  ftp_connection_free (conn);
 	  conn = NULL;
 	  g_mutex_lock (ftp->mutex);
 	  ftp->connections--;
-	  /* FIXME: This assignment is racy due to the mutex unlock above */
-	  ftp->max_connections = ftp->connections;
+	  continue;
+	}
+
+      if (ftp->connections < ftp->max_connections)
+	{
+	  /* Save current number of connections here, so we can limit maximum 
+	   * connections later.
+	   * This is necessary for threading reasons (connections can be 
+	   * opened or closed while we are still in the opening process. */
+	  guint maybe_max_connections = ftp->connections;
+
+	  ftp->connections++;
+	  g_mutex_unlock (ftp->mutex);
+	  conn = ftp_connection_create (ftp->addr, job);
+	  ftp_connection_prepare (conn);
+	  ftp_connection_login (conn, ftp->user, ftp->password);
+	  ftp_connection_use (conn);
+	  if (G_LIKELY (!ftp_connection_in_error (conn)))
+	    break;
+
+	  ftp_connection_clear_error (conn);
+	  /* Don't call ftp_connection_pop_job () here, the job isn't done yet */
+	  conn->job = NULL;
+	  ftp_connection_free (conn);
+	  conn = NULL;
+	  g_mutex_lock (ftp->mutex);
+	  ftp->connections--;
+	  ftp->max_connections = MIN (ftp->max_connections, maybe_max_connections);
 	  if (ftp->max_connections == 0)
 	    {
 	      DEBUG ("no more connections left, exiting...");
@@ -1294,6 +1447,8 @@ do_mount (GVfsBackend *backend,
       return;
     }
 
+  ftp_connection_prepare (conn);
+
   port = soup_address_get_port (ftp->addr);
   /* FIXME: need to translate this? */
   if (port == 21)
@@ -1333,9 +1488,15 @@ do_mount (GVfsBackend *backend,
     {
       GAskPasswordFlags flags;
       if (prompt == NULL)
-	/* translators: %s here is the hostname */
-	prompt = g_strdup_printf (_("Enter password for ftp on %s"), host);
-
+	{
+	  if (ftp->has_initial_user)
+	    /* Translators: the first %s is the username, the second the host name */
+	    prompt = g_strdup_printf (_("Enter password for ftp as %s on %s"), ftp->user, host);
+	  else
+	    /* translators: %s here is the hostname */
+	    prompt = g_strdup_printf (_("Enter password for ftp on %s"), host);
+	}
+	  
       flags = G_ASK_PASSWORD_NEED_PASSWORD;
         
       if (!ftp->has_initial_user)
@@ -1398,7 +1559,7 @@ try_login:
           !g_error_matches (conn->error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED))
 	break;
 
-      g_clear_error (&conn->error);
+      ftp_connection_clear_error (conn);
     }
   
   ftp_connection_use (conn);
@@ -2024,7 +2185,7 @@ resolve_symlink (GVfsBackendFtp *ftp, FtpConnection *conn, GFileInfo *original,
   if (ftp_connection_in_error (conn))
     {
       g_assert (info == NULL);
-      g_clear_error (&conn->error);
+      ftp_connection_clear_error (conn);
       return original;
     }
   if (info == NULL)
@@ -2174,7 +2335,7 @@ do_enumerate (GVfsBackend *backend,
       
       g_vfs_job_enumerate_done (job);
       conn->job = NULL;
-      g_clear_error (&conn->error);
+      ftp_connection_clear_error (conn);
     }
   else
     g_assert (files == NULL);
