@@ -36,6 +36,8 @@ struct OPAQUE_TYPE__GVfsBackendTrash
 
   TrashWatcher *watcher;
   TrashRoot *root;
+
+  guint thaw_timeout_id;
 };
 
 G_DEFINE_TYPE (GVfsBackendTrash, g_vfs_backend_trash, G_VFS_TYPE_BACKEND);
@@ -210,8 +212,8 @@ trash_backend_get_file (GVfsBackendTrash  *backend,
     }
 
   if (file == NULL)
-    g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
-                 _("No such file or directory"));
+    g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
+                         _("No such file or directory"));
 
   return file;
 }
@@ -226,8 +228,8 @@ trash_backend_open_for_read (GVfsBackend        *vfs_backend,
   GError *error = NULL;
 
   if (filename[1] == '\0')
-    g_set_error (&error, G_IO_ERROR, G_IO_ERROR_IS_DIRECTORY,
-                 _("Can't open directory"));
+    g_set_error_literal (&error, G_IO_ERROR, G_IO_ERROR_IS_DIRECTORY,
+                         _("Can't open directory"));
 
   else
     {
@@ -240,6 +242,7 @@ trash_backend_open_for_read (GVfsBackend        *vfs_backend,
           GFileInputStream *stream;
 
           stream = g_file_read (real, NULL, &error);
+          g_object_unref (real);
       
           if (stream)
             {
@@ -318,6 +321,7 @@ trash_backend_close_read (GVfsBackend       *backend,
   if (g_input_stream_close (handle, NULL, &error))
     {
       g_vfs_job_succeeded (G_VFS_JOB (job));
+      g_object_unref (handle);
 
       return TRUE;
     }
@@ -330,6 +334,28 @@ trash_backend_close_read (GVfsBackend       *backend,
   return TRUE;
 }
 
+static gboolean
+trash_backend_thaw_callback (gpointer user_data)
+{
+  GVfsBackendTrash *backend = user_data;
+
+  trash_root_thaw (backend->root);
+
+  backend->thaw_timeout_id = 0;
+  return FALSE;
+}
+
+static void
+trash_backend_schedule_thaw (GVfsBackendTrash *backend)
+{
+  if (backend->thaw_timeout_id)
+    g_source_remove (backend->thaw_timeout_id);
+
+  backend->thaw_timeout_id = g_timeout_add (200,
+                                            trash_backend_thaw_callback,
+                                            backend);
+}
+
 static gboolean
 trash_backend_delete (GVfsBackend   *vfs_backend,
                       GVfsJobDelete *job,
@@ -340,8 +366,8 @@ trash_backend_delete (GVfsBackend   *vfs_backend,
   g_print ("before job: %d\n", G_OBJECT(job)->ref_count);
 
   if (filename[1] == '\0')
-    g_set_error (&error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
-                 _("Can't delete trash"));
+    g_set_error_literal (&error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
+                         _("The trash folder may not be deleted"));
   else
     {
       gboolean is_toplevel;
@@ -357,13 +383,14 @@ trash_backend_delete (GVfsBackend   *vfs_backend,
           g_object_unref (real);
 
           if (!is_toplevel)
-            g_set_error (&error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
-                         _("Items in the trash may not be modified"));
+            g_set_error_literal (&error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
+                                 _("Items in the trash may not be modified"));
 
           else
             {
               if (trash_item_delete (item, &error))
                 {
+                  trash_backend_schedule_thaw (backend);
                   g_vfs_job_succeeded (G_VFS_JOB (job));
                   trash_item_unref (item);
 
@@ -395,8 +422,8 @@ trash_backend_pull (GVfsBackend           *vfs_backend,
   GError *error = NULL;
 
   if (source[1] == '\0')
-    g_set_error (&error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-                 _("Can't pull trash"));
+    g_set_error_literal (&error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
+                         _("The trash folder may not be deleted"));
   else
     {
       gboolean is_toplevel;
@@ -409,8 +436,8 @@ trash_backend_pull (GVfsBackend           *vfs_backend,
       if (real)
         {
           if (remove_source && !is_toplevel)
-            g_set_error (&error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
-                         _("Items in the trash may not be modified"));
+            g_set_error_literal (&error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
+                                 _("Items in the trash may not be modified"));
 
           else
             {
